@@ -1,4 +1,4 @@
-# Copyright 1999-2021 Gentoo Authors
+# Copyright 1999-2024 Gentoo Authors
 # Distributed under the terms of the GNU General Public License v2
 
 import functools
@@ -24,6 +24,7 @@ from portage.package.ebuild.prepare_build_dirs import (
     _prepare_fake_distdir,
     _prepare_fake_filesdir,
 )
+from portage.eapi import _get_eapi_attrs
 from portage.util import writemsg, ensure_dirs
 from portage.util._async.AsyncTaskFuture import AsyncTaskFuture
 from portage.util._async.BuildLogger import BuildLogger
@@ -54,20 +55,42 @@ portage.proxy.lazyimport.lazyimport(
     + "_post_src_install_write_metadata,"
     + "_preinst_bsdflags",
     "portage.util.futures.unix_events:_set_nonblocking",
+    "portage.util.locale:async_check_locale,split_LC_ALL",
 )
 from portage import os
 from portage import _encodings
 from portage import _unicode_encode
 
 
-class EbuildPhase(CompositeTask):
+async def _setup_locale(settings):
+    eapi_attrs = _get_eapi_attrs(settings["EAPI"])
+    if eapi_attrs.posixish_locale:
+        split_LC_ALL(settings)
+        settings["LC_COLLATE"] = "C"
+        # check_locale() returns None when check can not be executed.
+        if await async_check_locale(silent=True, env=settings.environ()) is False:
+            # try another locale
+            for l in ("C.UTF-8", "en_US.UTF-8", "en_GB.UTF-8", "C"):
+                settings["LC_CTYPE"] = l
+                if await async_check_locale(silent=True, env=settings.environ()):
+                    # TODO: output the following only once
+                    # writemsg(
+                    #     _("!!! LC_CTYPE unsupported, using %s instead\n")
+                    #     % self.settings["LC_CTYPE"]
+                    # )
+                    break
+            else:
+                raise AssertionError("C locale did not pass the test!")
 
+
+class EbuildPhase(CompositeTask):
     __slots__ = ("actionmap", "fd_pipes", "phase", "settings") + ("_ebuild_lock",)
 
     # FEATURES displayed prior to setup phase
     _features_display = (
         "ccache",
         "compressdebug",
+        "dedupdebug",
         "distcc",
         "fakeroot",
         "installsources",
@@ -96,11 +119,13 @@ class EbuildPhase(CompositeTask):
 
     async def _async_start(self):
 
+        await _setup_locale(self.settings)
+
         need_builddir = self.phase not in EbuildProcess._phases_without_builddir
 
         if need_builddir:
             phase_completed_file = os.path.join(
-                self.settings["PORTAGE_BUILDDIR"], ".%sed" % self.phase.rstrip("e")
+                self.settings["PORTAGE_BUILDDIR"], f".{self.phase.rstrip('e')}ed"
             )
             if not os.path.exists(phase_completed_file):
                 # If the phase is really going to run then we want
@@ -113,7 +138,6 @@ class EbuildPhase(CompositeTask):
             ensure_dirs(os.path.join(self.settings["PORTAGE_BUILDDIR"], "empty"))
 
         if self.phase in ("nofetch", "pretend", "setup"):
-
             use = self.settings.get("PORTAGE_BUILT_USE")
             if use is None:
                 use = self.settings["PORTAGE_USE"]
@@ -135,24 +159,22 @@ class EbuildPhase(CompositeTask):
                     maint_str = "<invalid metadata.xml>"
 
             msg = []
-            msg.append(
-                "Package:    {}:{}".format(self.settings.mycpv, self.settings["SLOT"])
-            )
+            msg.append(f"Package:    {self.settings.mycpv}:{self.settings['SLOT']}")
             if self.settings.get("PORTAGE_REPO_NAME"):
-                msg.append("Repository: %s" % self.settings["PORTAGE_REPO_NAME"])
+                msg.append(f"Repository: {self.settings['PORTAGE_REPO_NAME']}")
             if maint_str:
-                msg.append("Maintainer: %s" % maint_str)
+                msg.append(f"Maintainer: {maint_str}")
             if upstr_str:
-                msg.append("Upstream:   %s" % upstr_str)
+                msg.append(f"Upstream:   {upstr_str}")
 
-            msg.append("USE:        %s" % use)
+            msg.append(f"USE:        {use}")
             relevant_features = []
             enabled_features = self.settings.features
             for x in self._features_display:
                 if x in enabled_features:
                     relevant_features.append(x)
             if relevant_features:
-                msg.append("FEATURES:   %s" % " ".join(relevant_features))
+                msg.append(f"FEATURES:   {' '.join(relevant_features)}")
 
             # Force background=True for this header since it's intended
             # for the log and it doesn't necessarily need to be visible
@@ -395,7 +417,6 @@ class EbuildPhase(CompositeTask):
         self.wait()
 
     def _post_phase_exit(self, post_phase):
-
         self._assert_current(post_phase)
 
         log_path = None
@@ -408,7 +429,7 @@ class EbuildPhase(CompositeTask):
             self._append_temp_log(post_phase.logfile, log_path)
 
         if self._final_exit(post_phase) != os.EX_OK:
-            writemsg("!!! post %s failed; exiting.\n" % self.phase, noiselevel=-1)
+            writemsg(f"!!! post {self.phase} failed; exiting.\n", noiselevel=-1)
             self._die_hooks()
             return
 
@@ -417,7 +438,6 @@ class EbuildPhase(CompositeTask):
         return
 
     def _append_temp_log(self, temp_log, log_path):
-
         temp_file = open(
             _unicode_encode(temp_log, encoding=_encodings["fs"], errors="strict"), "rb"
         )
@@ -434,7 +454,6 @@ class EbuildPhase(CompositeTask):
         os.unlink(temp_log)
 
     def _open_log(self, log_path):
-
         f = open(
             _unicode_encode(log_path, encoding=_encodings["fs"], errors="strict"),
             mode="ab",
@@ -541,7 +560,6 @@ class EbuildPhase(CompositeTask):
 
 
 class _PostPhaseCommands(CompositeTask):
-
     __slots__ = ("commands", "elog", "fd_pipes", "logfile", "phase", "settings")
 
     def _start(self):
@@ -570,14 +588,13 @@ class _PostPhaseCommands(CompositeTask):
                     phase=self.phase,
                     scheduler=self.scheduler,
                     settings=self.settings,
-                    **kwargs
+                    **kwargs,
                 )
             )
 
         self._start_task(tasks, self._commands_exit)
 
     def _commands_exit(self, task):
-
         if self._default_exit(task) != os.EX_OK:
             self._async_wait()
             return
@@ -609,7 +626,6 @@ class _PostPhaseCommands(CompositeTask):
             self._default_final_exit(task)
 
     async def _soname_deps_qa(self):
-
         vardb = QueryCommand.get_db()[self.settings["EROOT"]]["vartree"].dbapi
 
         all_provides = await self.scheduler.run_in_executor(
@@ -625,7 +641,7 @@ class _PostPhaseCommands(CompositeTask):
             qa_msg = ["QA Notice: Unresolved soname dependencies:"]
             qa_msg.append("")
             qa_msg.extend(
-                "\t{}: {}".format(filename, " ".join(sorted(soname_deps)))
+                f"\t{filename}: {' '.join(sorted(soname_deps))}"
                 for filename, soname_deps in unresolved
             )
             qa_msg.append("")
