@@ -1,4 +1,4 @@
-# Copyright 1999-2021 Gentoo Authors
+# Copyright 1999-2024 Gentoo Authors
 # Distributed under the terms of the GNU General Public License v2
 
 from collections import deque
@@ -64,7 +64,6 @@ FAILURE = 1
 
 
 class Scheduler(PollScheduler):
-
     # max time between loadavg checks (seconds)
     _loadavg_latency = 30
 
@@ -340,7 +339,6 @@ class Scheduler(PollScheduler):
             )
 
     def _handle_self_update(self):
-
         if self._opts_no_self_update.intersection(self.myopts):
             return os.EX_OK
 
@@ -494,7 +492,6 @@ class Scheduler(PollScheduler):
         return interactive_tasks
 
     def _set_graph_config(self, graph_config):
-
         if graph_config is None:
             self._graph_config = None
             self._pkg_cache = {}
@@ -787,12 +784,10 @@ class Scheduler(PollScheduler):
         return os.EX_OK
 
     def _add_prefetchers(self):
-
         if not self._parallel_fetch:
             return
 
         if self._parallel_fetch:
-
             prefetchers = self._prefetchers
 
             for pkg in self._mergelist:
@@ -819,7 +814,6 @@ class Scheduler(PollScheduler):
             pass
 
         elif pkg.type_name == "ebuild":
-
             prefetcher = EbuildFetcher(
                 background=True,
                 config_pool=self._ConfigPool(
@@ -838,7 +832,6 @@ class Scheduler(PollScheduler):
             and "--getbinpkg" in self.myopts
             and pkg.root_config.trees["bintree"].isremote(pkg.cpv)
         ):
-
             prefetcher = BinpkgPrefetcher(
                 background=True, pkg=pkg, scheduler=self._sched_iface
             )
@@ -872,12 +865,13 @@ class Scheduler(PollScheduler):
             if self._terminated_tasks:
                 raise asyncio.CancelledError
 
-            out_str = "Running pre-merge checks for " + colorize("INFORM", x.cpv)
-            self._status_msg(out_str)
-
             root_config = x.root_config
             settings = self._allocate_config(root_config.root)
             settings.setcpv(x)
+
+            color = "PKG_BINARY_MERGE" if x.built else "INFORM"
+            self._status_msg(f"Running pre-merge checks for {colorize(color, x.cpv)}")
+
             if not x.built:
                 # Get required SRC_URI metadata (it's not cached in x.metadata
                 # because some packages have an extremely large SRC_URI value).
@@ -908,7 +902,6 @@ class Scheduler(PollScheduler):
             current_task = None
 
             try:
-
                 # Clean up the existing build dir, in case pkg_pretend
                 # checks for available space (bug #390711).
                 if existing_builddir:
@@ -937,6 +930,7 @@ class Scheduler(PollScheduler):
                     current_task = clean_phase
                     clean_phase.start()
                     await clean_phase.async_wait()
+                    current_task = None
 
                 if x.built:
                     tree = "bintree"
@@ -947,6 +941,10 @@ class Scheduler(PollScheduler):
                     # is consuming time here.
                     if bintree.isremote(x.cpv):
                         fetcher = self._get_prefetcher(x)
+                        if fetcher is not None and not fetcher.isAlive():
+                            # Cancel it because it hasn't started yet.
+                            fetcher.cancel()
+                            fetcher = None
                         if fetcher is None:
                             fetcher = BinpkgFetcher(pkg=x, scheduler=loop)
                             fetcher.start()
@@ -955,6 +953,16 @@ class Scheduler(PollScheduler):
                             # handles fetch, verification, and the
                             # bintree.inject call which moves the file.
                             fetched = fetcher.pkg_path
+                        else:
+                            msg = (
+                                "Fetching in the background:",
+                                fetcher.pkg_path,
+                                "To view fetch progress, run in another terminal:",
+                                f"tail -f {self._fetch_log}",
+                            )
+                            out = portage.output.EOutput()
+                            for l in msg:
+                                out.einfo(l)
                         if await fetcher.async_wait() != os.EX_OK:
                             failures += 1
                             self._record_pkg_failure(x, settings, fetcher.returncode)
@@ -974,16 +982,34 @@ class Scheduler(PollScheduler):
                         self._record_pkg_failure(x, settings, verifier.returncode)
                         continue
 
+                    current_task = None
                     if fetched:
-                        bintree.inject(
+                        if not bintree.inject(
                             x.cpv,
                             current_pkg_path=fetched,
                             allocated_pkg_path=fetcher.pkg_allocated_path,
-                        )
+                        ):
+                            eerror(
+                                "Binary package is not usable",
+                                phase="pretend",
+                                key=x.cpv,
+                            )
+                            failures += 1
+                            self._record_pkg_failure(x, settings, 1)
+                            continue
 
                     infloc = os.path.join(build_dir_path, "build-info")
                     ensure_dirs(infloc)
-                    await bintree.dbapi.unpack_metadata(settings, infloc, loop=loop)
+                    try:
+                        await bintree.dbapi.unpack_metadata(settings, infloc, loop=loop)
+                    except portage.exception.SignatureException as e:
+                        writemsg(
+                            f"!!! Invalid binary package: '{bintree.getname(x.cpv)}', {e}\n",
+                            noiselevel=-1,
+                        )
+                        failures += 1
+                        self._record_pkg_failure(x, settings, 1)
+                        continue
                     ebuild_path = os.path.join(infloc, x.pf + ".ebuild")
                     settings.configdict["pkg"]["EMERGE_FROM"] = "binary"
                     settings.configdict["pkg"]["MERGE_TYPE"] = "binary"
@@ -1023,24 +1049,27 @@ class Scheduler(PollScheduler):
                 current_task = pretend_phase
                 pretend_phase.start()
                 ret = await pretend_phase.async_wait()
+                # Leave current_task assigned in order to trigger clean
+                # on success in the below finally block.
                 if ret != os.EX_OK:
                     failures += 1
                     self._record_pkg_failure(x, settings, ret)
-                portage.elog.elog_process(x.cpv, settings)
             finally:
-
                 if current_task is not None:
                     if current_task.isAlive():
                         current_task.cancel()
-                    if current_task.returncode == os.EX_OK:
-                        clean_phase = EbuildPhase(
-                            background=False,
-                            phase="clean",
-                            scheduler=sched_iface,
-                            settings=settings,
-                        )
-                        clean_phase.start()
-                        await clean_phase.async_wait()
+
+                portage.elog.elog_process(x.cpv, settings)
+
+                if current_task is not None and current_task.returncode == os.EX_OK:
+                    clean_phase = EbuildPhase(
+                        background=False,
+                        phase="clean",
+                        scheduler=sched_iface,
+                        settings=settings,
+                    )
+                    clean_phase.start()
+                    await clean_phase.async_wait()
 
                 await build_dir.async_unlock()
                 self._deallocate_config(settings)
@@ -1126,7 +1155,6 @@ class Scheduler(PollScheduler):
             return rval
 
         while True:
-
             received_signal = []
 
             def sighandler(signum, frame):
@@ -1268,14 +1296,21 @@ class Scheduler(PollScheduler):
             and self._failed_pkgs_die_msgs
             and not mod_echo_output
         ):
+            failed_pkg_map = {}
+            for failed_pkg in self._failed_pkgs_all:
+                pkg = failed_pkg.pkg
+                failed_pkg_map[(pkg.cpv, pkg.root)] = pkg
 
             for mysettings, key, logentries in self._failed_pkgs_die_msgs:
+                pkg = failed_pkg_map.get((key, mysettings["EROOT"]))
+                color = "PKG_BINARY_MERGE" if pkg and pkg.built else "INFORM"
+
                 root_msg = ""
                 if mysettings["ROOT"] != "/":
                     root_msg = f" merged to {mysettings['ROOT']}"
                 print()
                 printer.einfo(
-                    f"Error messages for package {colorize('INFORM', key)}{root_msg}:"
+                    f"Error messages for package {colorize(color, key)}{root_msg}:"
                 )
                 print()
                 for phase in portage.const.EBUILD_PHASES:
@@ -1329,7 +1364,6 @@ class Scheduler(PollScheduler):
             self._failed_pkgs_die_msgs.append((mysettings, key, errors))
 
     def _locate_failure_log(self, failed_pkg):
-
         log_paths = [failed_pkg.build_log]
 
         for log_path in log_paths:
@@ -1487,14 +1521,16 @@ class Scheduler(PollScheduler):
             self.curval += 1
             merge = PackageMerge(merge=build, scheduler=self._sched_iface)
             self._running_tasks[id(merge)] = merge
-            if (
-                not build.build_opts.buildpkgonly
-                and build.pkg in self._deep_system_deps
+            # By default, merge-wait only allows merge when no builds are executing.
+            # As a special exception, dependencies on system packages are frequently
+            # unspecified and will therefore force merge-wait.
+            is_system_pkg = build.pkg in self._deep_system_deps
+            if not build.build_opts.buildpkgonly and (
+                "merge-wait" in build.settings.features or is_system_pkg
             ):
-                # Since dependencies on system packages are frequently
-                # unspecified, merge them only when no builds are executing.
                 self._merge_wait_queue.append(merge)
-                merge.addStartListener(self._system_merge_started)
+                if is_system_pkg:
+                    merge.addStartListener(self._system_merge_started)
             else:
                 self._task_queues.merge.add(merge)
                 merge.addExitListener(self._merge_exit)
@@ -1548,7 +1584,6 @@ class Scheduler(PollScheduler):
         self._event_loop.run_until_complete(self._main_exit)
 
     def _merge(self):
-
         if self._opts_no_background.intersection(self.myopts):
             self._set_max_jobs(1)
 
@@ -1755,9 +1790,7 @@ class Scheduler(PollScheduler):
         return self._jobs
 
     def _schedule_tasks(self):
-
         while True:
-
             state_change = 0
 
             # When the number of jobs and merges drops to zero,
@@ -1852,15 +1885,12 @@ class Scheduler(PollScheduler):
         """
 
         if self._jobs and self._max_load is not None:
-
             current_time = time.time()
 
             if self._sigcont_time is not None:
-
                 elapsed_seconds = current_time - self._sigcont_time
                 # elapsed_seconds < 0 means the system clock has been adjusted
                 if elapsed_seconds > 0 and elapsed_seconds < self._sigcont_delay:
-
                     if self._job_delay_timeout_id is not None:
                         self._job_delay_timeout_id.cancel()
 
@@ -1885,7 +1915,6 @@ class Scheduler(PollScheduler):
             elapsed_seconds = current_time - self._previous_job_start_time
             # elapsed_seconds < 0 means the system clock has been adjusted
             if elapsed_seconds > 0 and elapsed_seconds < delay:
-
                 if self._job_delay_timeout_id is not None:
                     self._job_delay_timeout_id.cancel()
 
@@ -1905,7 +1934,6 @@ class Scheduler(PollScheduler):
         state_change = 0
 
         while True:
-
             if not self._keep_scheduling():
                 return bool(state_change)
 
@@ -1964,7 +1992,9 @@ class Scheduler(PollScheduler):
             # CPython 2.7, so it may be possible for CPython to raise KeyError
             # here as well.
             prefetcher = None
-        if prefetcher is not None and not prefetcher.isAlive():
+        if prefetcher is not None and (
+            prefetcher.cancelled or not prefetcher.isAlive()
+        ):
             try:
                 self._task_queues.fetch._task_queue.remove(prefetcher)
             except ValueError:
@@ -1973,7 +2003,6 @@ class Scheduler(PollScheduler):
         return prefetcher
 
     def _task(self, pkg):
-
         pkg_to_replace = None
         if pkg.operation != "uninstall":
             vardb = pkg.root_config.trees["vartree"].dbapi
@@ -2021,7 +2050,10 @@ class Scheduler(PollScheduler):
 
     def _failed_pkg_msg(self, failed_pkg, action, preposition):
         pkg = failed_pkg.pkg
-        msg = f"{bad('Failed')} to {action} {colorize('INFORM', pkg.cpv)}"
+
+        color = "PKG_BINARY_MERGE" if failed_pkg.pkg.built else "INFORM"
+
+        msg = f"{bad('Failed')} to {action} {colorize(color, pkg.cpv)}"
         if pkg.root_config.settings["ROOT"] != "/":
             msg += f" {preposition} {pkg.root}"
 
@@ -2220,7 +2252,6 @@ class Scheduler(PollScheduler):
             atom = self._world_atoms.get(pkg)
 
         try:
-
             if hasattr(world_set, "lock"):
                 world_set.lock()
                 world_locked = True
@@ -2240,11 +2271,11 @@ class Scheduler(PollScheduler):
                 if atom is not None:
                     if hasattr(world_set, "add"):
                         self._status_msg(
-                            ('Recording %s in "world" ' + "favorites file...") % atom
+                            f'Recording {atom} in "world" favorites file...'
                         )
                         logger.log(
-                            " === (%s of %s) Updating world file (%s)"
-                            % (pkg_count.curval, pkg_count.maxval, pkg.cpv)
+                            f" === ({pkg_count.curval} of {pkg_count.maxval}) "
+                            f"Updating world file ({pkg.cpv})"
                         )
                         world_set.add(atom)
                     else:
