@@ -1,4 +1,4 @@
-# Copyright 2008-2021 Gentoo Authors
+# Copyright 2008-2024 Gentoo Authors
 # Distributed under the terms of the GNU General Public License v2
 
 import functools
@@ -16,7 +16,6 @@ from portage.util.futures import asyncio
 
 
 class SpawnProcess(SubProcess):
-
     """
     Constructor keyword args are passed into portage.process.spawn().
     The required "args" keyword argument will be passed as the first
@@ -42,7 +41,7 @@ class SpawnProcess(SubProcess):
     )
 
     __slots__ = (
-        ("args", "log_filter_file")
+        ("args", "create_pipe", "log_filter_file")
         + _spawn_kwarg_names
         + (
             "_main_task",
@@ -61,56 +60,78 @@ class SpawnProcess(SubProcess):
         else:
             self.fd_pipes = self.fd_pipes.copy()
         fd_pipes = self.fd_pipes
+        log_file_path = None
 
-        master_fd, slave_fd = self._pipe(fd_pipes)
+        if fd_pipes or self.logfile or not self.background:
+            if self.create_pipe is not False:
+                master_fd, slave_fd = self._pipe(fd_pipes)
 
-        can_log = self._can_log(slave_fd)
-        if can_log:
-            log_file_path = self.logfile
+                can_log = self._can_log(slave_fd)
+                if can_log:
+                    log_file_path = self.logfile
+            else:
+                if self.logfile:
+                    raise NotImplementedError(
+                        "logfile conflicts with create_pipe=False"
+                    )
+                # When called via process.spawn and ForkProcess._start,
+                # SpawnProcess will have created a pipe earlier, so it
+                # would be redundant to do it here (it could also trigger
+                # spawn recursion via set_term_size as in bug 923750).
+                # Use /dev/null for master_fd, triggering early return
+                # of _main, followed by _async_waitpid.
+                # TODO: Optimize away the need for master_fd here.
+                master_fd = os.open(os.devnull, os.O_RDONLY)
+                slave_fd = None
+                can_log = False
+
+            null_input = None
+            if not self.background or 0 in fd_pipes:
+                # Subclasses such as AbstractEbuildProcess may have already passed
+                # in a null file descriptor in fd_pipes, so use that when given.
+                pass
+            else:
+                # TODO: Use job control functions like tcsetpgrp() to control
+                # access to stdin. Until then, use /dev/null so that any
+                # attempts to read from stdin will immediately return EOF
+                # instead of blocking indefinitely.
+                null_input = os.open("/dev/null", os.O_RDWR)
+                fd_pipes[0] = null_input
+
+            fd_pipes.setdefault(0, portage._get_stdin().fileno())
+            fd_pipes.setdefault(1, sys.__stdout__.fileno())
+            fd_pipes.setdefault(2, sys.__stderr__.fileno())
+
+            # flush any pending output
+            stdout_filenos = (sys.__stdout__.fileno(), sys.__stderr__.fileno())
+            for fd in fd_pipes.values():
+                if fd in stdout_filenos:
+                    sys.__stdout__.flush()
+                    sys.__stderr__.flush()
+                    break
+
+            fd_pipes_orig = fd_pipes.copy()
+
+            if slave_fd is None:
+                pass
+            elif log_file_path is not None or self.background:
+                fd_pipes[1] = slave_fd
+                fd_pipes[2] = slave_fd
+
+            else:
+                # Create a dummy pipe that PipeLogger uses to efficiently
+                # monitor for process exit by listening for the EOF event.
+                # Re-use of the allocated fd number for the key in fd_pipes
+                # guarantees that the keys will not collide for similarly
+                # allocated pipes which are used by callers such as
+                # FileDigester and MergeProcess. See the _setup_pipes
+                # docstring for more benefits of this allocation approach.
+                self._dummy_pipe_fd = slave_fd
+                fd_pipes[slave_fd] = slave_fd
         else:
-            log_file_path = None
-
-        null_input = None
-        if not self.background or 0 in fd_pipes:
-            # Subclasses such as AbstractEbuildProcess may have already passed
-            # in a null file descriptor in fd_pipes, so use that when given.
-            pass
-        else:
-            # TODO: Use job control functions like tcsetpgrp() to control
-            # access to stdin. Until then, use /dev/null so that any
-            # attempts to read from stdin will immediately return EOF
-            # instead of blocking indefinitely.
-            null_input = os.open("/dev/null", os.O_RDWR)
-            fd_pipes[0] = null_input
-
-        fd_pipes.setdefault(0, portage._get_stdin().fileno())
-        fd_pipes.setdefault(1, sys.__stdout__.fileno())
-        fd_pipes.setdefault(2, sys.__stderr__.fileno())
-
-        # flush any pending output
-        stdout_filenos = (sys.__stdout__.fileno(), sys.__stderr__.fileno())
-        for fd in fd_pipes.values():
-            if fd in stdout_filenos:
-                sys.__stdout__.flush()
-                sys.__stderr__.flush()
-                break
-
-        fd_pipes_orig = fd_pipes.copy()
-
-        if log_file_path is not None or self.background:
-            fd_pipes[1] = slave_fd
-            fd_pipes[2] = slave_fd
-
-        else:
-            # Create a dummy pipe that PipeLogger uses to efficiently
-            # monitor for process exit by listening for the EOF event.
-            # Re-use of the allocated fd number for the key in fd_pipes
-            # guarantees that the keys will not collide for similarly
-            # allocated pipes which are used by callers such as
-            # FileDigester and MergeProcess. See the _setup_pipes
-            # docstring for more benefits of this allocation approach.
-            self._dummy_pipe_fd = slave_fd
-            fd_pipes[slave_fd] = slave_fd
+            can_log = False
+            slave_fd = None
+            null_input = None
 
         kwargs = {}
         for k in self._spawn_kwarg_names:
@@ -119,27 +140,31 @@ class SpawnProcess(SubProcess):
                 kwargs[k] = v
 
         kwargs["fd_pipes"] = fd_pipes
-        kwargs["returnpid"] = True
+        kwargs["returnproc"] = True
         kwargs.pop("logfile", None)
 
-        retval = self._spawn(self.args, **kwargs)
+        self._proc = self._spawn(self.args, **kwargs)
 
-        os.close(slave_fd)
+        if slave_fd is not None:
+            os.close(slave_fd)
         if null_input is not None:
             os.close(null_input)
 
-        if isinstance(retval, int):
-            # spawn failed
-            self.returncode = retval
-            self._async_wait()
+        if not fd_pipes:
+            self._registered = True
+            self._async_waitpid()
             return
 
-        self.pid = retval[0]
-
         stdout_fd = None
         if can_log and not self.background:
             stdout_fd = os.dup(fd_pipes_orig[1])
 
+        self._start_main_task(
+            master_fd, log_file_path=log_file_path, stdout_fd=stdout_fd
+        )
+        self._registered = True
+
+    def _start_main_task(self, pr, log_file_path=None, stdout_fd=None):
         build_logger = BuildLogger(
             env=self.env,
             log_path=log_file_path,
@@ -151,14 +176,13 @@ class SpawnProcess(SubProcess):
         pipe_logger = PipeLogger(
             background=self.background,
             scheduler=self.scheduler,
-            input_fd=master_fd,
+            input_fd=pr,
             log_file_path=build_logger.stdin,
             stdout_fd=stdout_fd,
         )
 
         pipe_logger.start()
 
-        self._registered = True
         self._main_task_cancel = functools.partial(
             self._main_cancel, build_logger, pipe_logger
         )
@@ -217,7 +241,9 @@ class SpawnProcess(SubProcess):
         got_pty, master_fd, slave_fd = _create_pty_or_pipe(copy_term_size=stdout_pipe)
         return (master_fd, slave_fd)
 
-    def _spawn(self, args, **kwargs):
+    def _spawn(
+        self, args: list[str], **kwargs
+    ) -> portage.process.MultiprocessingProcess:
         spawn_func = portage.process.spawn
 
         if self._selinux_type is not None:
