@@ -1,4 +1,4 @@
-# Copyright 1998-2021 Gentoo Authors
+# Copyright 1998-2024 Gentoo Authors
 # Distributed under the terms of the GNU General Public License v2
 
 __all__ = ["bindbapi", "binarytree"]
@@ -37,6 +37,7 @@ from portage.dbapi.virtual import fakedbapi
 from portage.dep import Atom, use_reduce, paren_enclose
 from portage.exception import (
     AlarmSignal,
+    CorruptionKeyError,
     InvalidPackageName,
     InvalidBinaryPackageFormat,
     ParseError,
@@ -84,9 +85,12 @@ class bindbapi(fakedbapi):
     _known_keys = frozenset(
         list(fakedbapi._known_keys) + ["CHOST", "repository", "USE"]
     )
+    # Must include keys used to create _pkg_str attributes used in
+    # the fakedbapi _instance_key_multi_instance method.
     _pkg_str_aux_keys = fakedbapi._pkg_str_aux_keys + (
         "BUILD_ID",
         "BUILD_TIME",
+        "SIZE",
         "_mtime_",
     )
 
@@ -129,8 +133,29 @@ class bindbapi(fakedbapi):
             "USE",
             "_mtime_",
         }
-        self._aux_cache_slot_dict = slot_dict_class(self._aux_cache_keys)
         self._aux_cache = {}
+        self._aux_cache_slot_dict_cache = None
+
+    @property
+    def _aux_cache_slot_dict(self):
+        if self._aux_cache_slot_dict_cache is None:
+            self._aux_cache_slot_dict_cache = slot_dict_class(self._aux_cache_keys)
+        return self._aux_cache_slot_dict_cache
+
+    def __getstate__(self):
+        state = self.__dict__.copy()
+        # These attributes are not picklable, so they are automatically
+        # regenerated after unpickling.
+        state["_aux_cache_slot_dict_cache"] = None
+        state["_instance_key"] = None
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+        if self._multi_instance:
+            self._instance_key = self._instance_key_multi_instance
+        else:
+            self._instance_key = self._instance_key_cpv
 
     @property
     def writable(self):
@@ -186,9 +211,9 @@ class bindbapi(fakedbapi):
                 raise KeyError(mycpv)
             binpkg_path = os.path.join(self.bintree.pkgdir, binpkg_path)
             try:
-                st = os.lstat(binpkg_path)
-            except OSError:
-                raise KeyError(mycpv)
+                st = os.stat(binpkg_path)
+            except OSError as oe:
+                raise CorruptionKeyError(mycpv) from oe
 
             binpkg_format = get_binpkg_format(binpkg_path)
             if binpkg_format == "xpak":
@@ -259,8 +284,10 @@ class bindbapi(fakedbapi):
             cpv_str += f"-{build_id}"
 
         binpkg_path = self.bintree.getname(cpv)
-        if not os.path.exists(binpkg_path):
-            raise KeyError(cpv)
+        try:
+            os.stat(binpkg_path)
+        except OSError as oe:
+            raise CorruptionKeyError(cpv) from oe
 
         binpkg_format = get_binpkg_format(binpkg_path)
         if binpkg_format == "xpak":
@@ -269,11 +296,25 @@ class bindbapi(fakedbapi):
             encoding_key = True
         elif binpkg_format == "gpkg":
             mybinpkg = portage.gpkg.gpkg(self.settings, cpv_str, binpkg_path)
-            mydata = mybinpkg.get_metadata()
+            try:
+                mydata = mybinpkg.get_metadata()
+                signature_exist = mybinpkg.signature_exist
+            except SignatureException:
+                signature_exist = True
+            if signature_exist:
+                writemsg(
+                    colorize(
+                        "WARN",
+                        f"Binpkg update ignored for signed package: {binpkg_path}, "
+                        "the file will be removed.\n",
+                    )
+                )
+                self.bintree.remove(cpv)
+                return
             encoding_key = False
         else:
             raise InvalidBinaryPackageFormat(
-                "Unknown binary package format %s" % binpkg_path
+                f"Unknown binary package format {binpkg_path}"
             )
 
         for k, v in values.items():
@@ -295,7 +336,7 @@ class bindbapi(fakedbapi):
             mybinpkg.update_metadata(mydata)
         else:
             raise InvalidBinaryPackageFormat(
-                "Unknown binary package format %s" % binpkg_path
+                f"Unknown binary package format {binpkg_path}"
             )
 
         # inject will clear stale caches via cpv_inject.
@@ -336,7 +377,7 @@ class bindbapi(fakedbapi):
                 )
             else:
                 raise InvalidBinaryPackageFormat(
-                    "Unknown binary package format %s" % binpkg_file
+                    f"Unknown binary package format {binpkg_file}"
                 )
 
     async def unpack_contents(self, pkg, dest_dir, loop=None):
@@ -374,7 +415,7 @@ class bindbapi(fakedbapi):
                 extractor.start()
                 await extractor.async_wait()
                 if extractor.returncode != os.EX_OK:
-                    raise PortageException("Error Extracting '{}'".format(pkg_path))
+                    raise PortageException(f"Error Extracting '{pkg_path}'")
             elif binpkg_format == "gpkg":
                 await loop.run_in_executor(
                     ForkExecutor(loop=loop),
@@ -426,7 +467,7 @@ class bindbapi(fakedbapi):
             except KeyError:
                 raise portage.exception.MissingSignature("SIZE")
             except ValueError:
-                raise portage.exception.InvalidSignature("SIZE: %s" % metadata["SIZE"])
+                raise portage.exception.InvalidSignature(f"SIZE: {metadata['SIZE']}")
             else:
                 filesdict[os.path.basename(self.bintree.getname(pkg))] = size
 
@@ -443,7 +484,6 @@ class binarytree:
         virtual=DeprecationWarning,
         settings=None,
     ):
-
         if pkgdir is None:
             raise TypeError("pkgdir parameter is required")
 
@@ -486,6 +526,7 @@ class binarytree:
         self._remotepkgs = None  # remote metadata indexed by cpv
         self._additional_pkgs = {}
         self.invalids = []
+        self.invalid_paths: dict[str, list[str]] = {}
         self.settings = settings
         self._pkg_paths = {}
         self._populating = False
@@ -644,7 +685,7 @@ class binarytree:
             # If this update has already been applied to the same
             # package build then silently continue.
             applied = False
-            for maybe_applied in self.dbapi.match("={}".format(mynewcpv)):
+            for maybe_applied in self.dbapi.match(f"={mynewcpv}"):
                 if maybe_applied.build_time == mycpv.build_time:
                     applied = True
                     break
@@ -660,13 +701,23 @@ class binarytree:
                 continue
 
             binpkg_path = self.getname(mycpv)
-            if os.path.exists(binpkg_path) and not os.access(binpkg_path, os.W_OK):
+            try:
+                os.stat(binpkg_path)
+            except FileNotFoundError:
+                writemsg(_("!!! File not found: %s\n") % binpkg_path, noiselevel=-1)
+                continue
+            except OSError as oe:
+                writemsg(
+                    _("!!! File os error (path %s): %s\n") % (binpkg_path, oe),
+                    noiselevel=-1,
+                )
+                continue
+            if not os.access(binpkg_path, os.W_OK):
                 writemsg(
                     _("!!! Cannot update readonly binary: %s\n") % mycpv, noiselevel=-1
                 )
                 continue
 
-            moves += 1
             binpkg_format = get_binpkg_format(binpkg_path)
             if binpkg_format == "xpak":
                 mytbz2 = portage.xpak.tbz2(binpkg_path)
@@ -674,11 +725,25 @@ class binarytree:
                 decode_metadata_name = False
             elif binpkg_format == "gpkg":
                 mybinpkg = portage.gpkg.gpkg(self.settings, mycpv, binpkg_path)
-                mydata = mybinpkg.get_metadata()
+                try:
+                    mydata = mybinpkg.get_metadata()
+                    signature_exist = mybinpkg.signature_exist
+                except SignatureException:
+                    signature_exist = True
+                if signature_exist:
+                    writemsg(
+                        colorize(
+                            "WARN",
+                            f"Binpkg update ignored for signed package: {binpkg_path}\n",
+                        )
+                    )
+                    continue
                 decode_metadata_name = True
             else:
                 continue
 
+            moves += 1
+
             updated_items = update_dbentries([mylist], mydata, parent=mycpv)
             mydata.update(updated_items)
             if decode_metadata_name:
@@ -795,7 +860,15 @@ class binarytree:
             except PortageException:
                 pass
 
-    def populate(self, getbinpkgs=False, getbinpkg_refresh=False, add_repos=()):
+    def populate(
+        self,
+        getbinpkgs=False,
+        getbinpkg_refresh=False,
+        add_repos=(),
+        force_reindex=False,
+        invalid_errors=True,
+        pretend=False,
+    ):
         """
         Populates the binarytree with package metadata.
 
@@ -808,6 +881,10 @@ class binarytree:
         @type add_repos: sequence
         """
 
+        # TODO: Should we return here if we're --pretend? On the one hand,
+        # people might not want --pretend to affect state. On the other hand,
+        # it makes --pretend pretty useless with --getbinpkg as your index will
+        # be stale.
         if self._populating:
             return
 
@@ -825,6 +902,8 @@ class binarytree:
         try:
             update_pkgindex = self._populate_local(
                 reindex="pkgdir-index-trusted" not in self.settings.features
+                or force_reindex,
+                invalid_errors=invalid_errors,
             )
 
             if update_pkgindex and self.dbapi.writable:
@@ -860,14 +939,16 @@ class binarytree:
                         noiselevel=-1,
                     )
                 else:
-                    self._populate_remote(getbinpkg_refresh=getbinpkg_refresh)
+                    self._populate_remote(
+                        getbinpkg_refresh=getbinpkg_refresh, pretend=pretend
+                    )
 
         finally:
             self._populating = False
 
         self.populated = True
 
-    def _populate_local(self, reindex=True):
+    def _populate_local(self, reindex=True, invalid_errors=True):
         """
         Populates the binarytree with local package metadata.
 
@@ -1004,18 +1085,24 @@ class binarytree:
                             self.dbapi.cpv_inject(mycpv)
                             continue
                     if not os.access(full_path, os.R_OK):
-                        writemsg(
-                            _("!!! Permission denied to read " "binary package: '%s'\n")
-                            % full_path,
-                            noiselevel=-1,
-                        )
+                        if invalid_errors:
+                            writemsg(
+                                _(
+                                    "!!! Permission denied to read "
+                                    "binary package: '%s'\n"
+                                )
+                                % full_path,
+                                noiselevel=-1,
+                            )
                         self.invalids.append(myfile[:-5])
+                        self.invalid_paths[myfile] = [full_path]
                         continue
 
                     try:
                         binpkg_format = get_binpkg_format(myfile)
                     except InvalidBinaryPackageFormat:
                         self.invalids.append(myfile[:-5])
+                        self.invalid_paths[myfile[:-5]] = [full_path]
                         continue
 
                     if gpkg_only:
@@ -1033,6 +1120,10 @@ class binarytree:
                         else:
                             binpkg_format = "gpkg"
 
+                    for ext in SUPPORTED_XPAK_EXTENSIONS + SUPPORTED_GPKG_EXTENSIONS:
+                        if myfile.endswith(ext):
+                            mypkg = myfile[: -len(ext)]
+                            break
                     try:
                         pkg_metadata = self._read_metadata(
                             full_path,
@@ -1041,24 +1132,23 @@ class binarytree:
                             binpkg_format=binpkg_format,
                         )
                     except (PortagePackageException, SignatureException) as e:
-                        writemsg(
-                            f"!!! Invalid binary package: '{full_path}', {e}\n",
-                            noiselevel=-1,
-                        )
+                        if invalid_errors:
+                            writemsg(
+                                f"!!! Invalid binary package: '{full_path}', {e}\n",
+                                noiselevel=-1,
+                            )
+                        self.invalid_paths[mypkg] = [full_path]
                         continue
                     mycat = pkg_metadata.get("CATEGORY", "")
                     mypf = pkg_metadata.get("PF", "")
                     slot = pkg_metadata.get("SLOT", "")
-                    for ext in SUPPORTED_XPAK_EXTENSIONS + SUPPORTED_GPKG_EXTENSIONS:
-                        if myfile.endswith(ext):
-                            mypkg = myfile[: -len(ext)]
-                            break
                     if not mycat or not mypf or not slot:
                         # old-style or corrupt package
-                        writemsg(
-                            _("\n!!! Invalid binary package: '%s'\n") % full_path,
-                            noiselevel=-1,
-                        )
+                        if invalid_errors:
+                            writemsg(
+                                _("\n!!! Invalid binary package: '%s'\n") % full_path,
+                                noiselevel=-1,
+                            )
                         missing_keys = []
                         if not mycat:
                             missing_keys.append("CATEGORY")
@@ -1066,22 +1156,25 @@ class binarytree:
                             missing_keys.append("PF")
                         if not slot:
                             missing_keys.append("SLOT")
-                        msg = []
-                        if missing_keys:
-                            missing_keys.sort()
+                        if invalid_errors:
+                            msg = []
+                            if missing_keys:
+                                missing_keys.sort()
+                                msg.append(
+                                    _("Missing metadata key(s): %s.")
+                                    % ", ".join(missing_keys)
+                                )
+                        if invalid_errors:
                             msg.append(
-                                _("Missing metadata key(s): %s.")
-                                % ", ".join(missing_keys)
+                                _(
+                                    " This binary package is not "
+                                    "recoverable and should be deleted."
+                                )
                             )
-                        msg.append(
-                            _(
-                                " This binary package is not "
-                                "recoverable and should be deleted."
-                            )
-                        )
-                        for line in textwrap.wrap("".join(msg), 72):
-                            writemsg("!!! %s\n" % line, noiselevel=-1)
+                            for line in textwrap.wrap("".join(msg), 72):
+                                writemsg(f"!!! {line}\n", noiselevel=-1)
                         self.invalids.append(mypkg)
+                        self.invalid_paths[mypkg] = [full_path]
                         continue
 
                     multi_instance = False
@@ -1092,7 +1185,7 @@ class binarytree:
                         build_id = self._parse_build_id(myfile)
                         if build_id < 1:
                             invalid_name = True
-                        elif myfile != "{}-{}.xpak".format(mypf, build_id):
+                        elif myfile != f"{mypf}-{build_id}.xpak":
                             invalid_name = True
                         else:
                             mypkg = mypkg[: -len(str(build_id)) - 1]
@@ -1100,12 +1193,12 @@ class binarytree:
                         build_id = self._parse_build_id(myfile)
                         if build_id > 0:
                             multi_instance = True
-                            if myfile != "{}-{}.gpkg.tar".format(mypf, build_id):
+                            if myfile != f"{mypf}-{build_id}.gpkg.tar":
                                 invalid_name = True
                             else:
                                 mypkg = mypkg[: -len(str(build_id)) - 1]
                         else:
-                            if myfile != "%s.gpkg.tar" % mypf:
+                            if myfile != f"{mypf}.gpkg.tar":
                                 invalid_name = True
                     elif myfile != mypf + ".tbz2":
                         invalid_name = True
@@ -1131,7 +1224,7 @@ class binarytree:
                         build_id = None
 
                     if multi_instance:
-                        name_split = catpkgsplit("{}/{}".format(mycat, mypf))
+                        name_split = catpkgsplit(f"{mycat}/{mypf}")
                         if (
                             name_split is None
                             or tuple(catsplit(mydir)) != name_split[:2]
@@ -1223,12 +1316,34 @@ class binarytree:
 
         return pkgindex if update_pkgindex else None
 
-    def _populate_remote(self, getbinpkg_refresh=True):
+    def _run_trust_helper(self):
+        portage_trust_helper = self.settings.get("PORTAGE_TRUST_HELPER", "")
+        if portage_trust_helper == "":
+            return
+        try:
+            ret = subprocess.run(portage_trust_helper)
+        except FileNotFoundError:
+            writemsg(
+                _(
+                    "\n!!! Portage trust helper %s for binary packages not found\n!!! Continuing, but did you install app-portage/getuto?\n"
+                )
+                % portage_trust_helper,
+                noiselevel=-1,
+            )
+            return
+        ret.check_returncode()
 
+    def _populate_remote(self, getbinpkg_refresh=True, pretend=False):
         self._remote_has_index = False
         self._remotepkgs = {}
 
         if "binpkg-request-signature" in self.settings.features:
+            # This is somewhat broken, we *should* run the trust helper always
+            # when binpackages are involved, not only when we refuse unsigned
+            # ones. (If the keys have expired we end up refusing signed but
+            # technically invalid packages...)
+            if not pretend and self.dbapi.writable:
+                self._run_trust_helper()
             gpkg_only = True
         else:
             gpkg_only = False
@@ -1237,23 +1352,13 @@ class binarytree:
         for repo in reversed(list(self._binrepos_conf.values())):
             base_url = repo.sync_uri
             parsed_url = urlparse(base_url)
-            host = parsed_url.netloc
+            host = parsed_url.hostname or ""
             port = parsed_url.port
-            user = None
-            passwd = None
-            user_passwd = ""
+            user = parsed_url.username
+            passwd = parsed_url.password
+            user_passwd = user + "@" if user else ""
             gpkg_only_warned = False
 
-            if "@" in host:
-                user, host = host.split("@", 1)
-                user_passwd = user + "@"
-                if ":" in user:
-                    user, passwd = user.split(":", 1)
-
-            if port is not None:
-                port_str = ":{}".format(port)
-                if host.endswith(port_str):
-                    host = host[: -len(port_str)]
             pkgindex_file = os.path.join(
                 self.settings["EROOT"],
                 CACHE_PATH,
@@ -1318,15 +1423,18 @@ class binarytree:
 
                 # Don't use urlopen for https, unless
                 # PEP 476 is supported (bug #469888).
-                if repo.fetchcommand is None and (
-                    parsed_url.scheme not in ("https",) or _have_pep_476()
-                ):
+                if (
+                    repo.fetchcommand is None or parsed_url.scheme in ("", "file")
+                ) and (parsed_url.scheme not in ("https",) or _have_pep_476()):
                     try:
-                        f = _urlopen(
-                            url, if_modified_since=local_timestamp, proxies=proxies
-                        )
-                        if hasattr(f, "headers") and f.headers.get("timestamp", ""):
-                            remote_timestamp = f.headers.get("timestamp")
+                        if parsed_url.scheme in ("", "file"):
+                            f = open(f"{parsed_url.path.rstrip('/')}/Packages", "rb")
+                        else:
+                            f = _urlopen(
+                                url, if_modified_since=local_timestamp, proxies=proxies
+                            )
+                            if hasattr(f, "headers") and f.headers.get("timestamp", ""):
+                                remote_timestamp = f.headers.get("timestamp")
                     except OSError as err:
                         if (
                             hasattr(err, "code") and err.code == 304
@@ -1343,11 +1451,10 @@ class binarytree:
                             raise
                     except ValueError:
                         raise ParseError(
-                            "Invalid Portage BINHOST value '%s'" % url.lstrip()
+                            f"Invalid Portage BINHOST value '{url.lstrip()}'"
                         )
 
                 if f is None:
-
                     path = parsed_url.path.rstrip("/") + "/Packages"
 
                     if repo.fetchcommand is None and parsed_url.scheme == "ssh":
@@ -1356,7 +1463,7 @@ class binarytree:
                         # matches that of the cached Packages file.
                         ssh_args = ["ssh"]
                         if port is not None:
-                            ssh_args.append("-p{}".format(port))
+                            ssh_args.append(f"-p{port}")
                         # NOTE: shlex evaluates embedded quotes
                         ssh_args.extend(
                             portage.util.shlex_split(
@@ -1400,7 +1507,7 @@ class binarytree:
                             fcmd=fcmd, fcmd_vars=fcmd_vars
                         )
                         if not success:
-                            raise OSError("{} failed".format(setting))
+                            raise OSError(f"{setting} failed")
                         f = open(tmp_filename, "rb")
 
                 f_dec = codecs.iterdecode(
@@ -1475,7 +1582,7 @@ class binarytree:
                     error_msg = str(e)
                 except UnicodeDecodeError as uerror:
                     error_msg = str(uerror.object, encoding="utf_8", errors="replace")
-                writemsg("!!! %s\n\n" % error_msg)
+                writemsg(f"!!! {error_msg}\n\n")
                 del e
                 pkgindex = None
             if proc is not None:
@@ -1689,6 +1796,57 @@ class binarytree:
 
         return cpv
 
+    def remove(self, cpv: portage.versions._pkg_str) -> None:
+        """
+        Remove a package instance and update internal state including
+        the package index. This will raise a KeyError if cpv is not
+        found in the internal state. It will display a warning message
+        if the package file was not found on disk, since it could have
+        been removed by another process before this method could
+        acquire a lock.
+
+        @param cpv: The cpv of the existing package to remove
+        @type cpv: portage.versions._pkg_str
+        @rtype: None
+        @return: None
+        @raise KeyError: If cpv does not exist in the internal state
+        """
+        if not self.populated:
+            self.populate()
+        os.makedirs(self.pkgdir, exist_ok=True)
+        pkgindex_lock = lockfile(self._pkgindex_file, wantnewlockfile=1)
+        try:
+            # Will raise KeyError if the package is not found.
+            instance_key = self.dbapi._instance_key(cpv)
+            pkg_path = self.getname(cpv)
+            self.dbapi.cpv_remove(cpv)
+            self._pkg_paths.pop(instance_key, None)
+            if self._remotepkgs is not None:
+                self._remotepkgs.pop(instance_key, None)
+            pkgindex = self._load_pkgindex()
+            if not self._pkgindex_version_supported(pkgindex):
+                pkgindex = self._new_pkgindex()
+
+            path = pkg_path[len(self.pkgdir) + 1 :]
+            for i in range(len(pkgindex.packages) - 1, -1, -1):
+                d = pkgindex.packages[i]
+                if cpv == d.get("CPV"):
+                    if path == d.get("PATH", ""):
+                        del pkgindex.packages[i]
+
+            self._pkgindex_write(pkgindex)
+            try:
+                os.remove(pkg_path)
+            except OSError as err:
+                writemsg(
+                    colorize(
+                        "WARN",
+                        f"Failed to remove package: {pkg_path} {str(err)}",
+                    )
+                )
+        finally:
+            unlockfile(pkgindex_lock)
+
     def _read_metadata(self, filename, st, keys=None, binpkg_format=None):
         """
         Read metadata from a binary package. The returned metadata
@@ -1977,7 +2135,7 @@ class binarytree:
                 deps = use_reduce(deps, uselist=use, token_class=token_class)
                 deps = paren_enclose(deps)
             except portage.exception.InvalidDependString as e:
-                writemsg("{}: {}\n".format(k, e), noiselevel=-1)
+                writemsg(f"{k}: {e}\n", noiselevel=-1)
                 raise
             metadata[k] = deps
 
@@ -1993,13 +2151,13 @@ class binarytree:
         if not self.populated:
             self.populate()
         writemsg("\n\n", 1)
-        writemsg("mydep: %s\n" % mydep, 1)
+        writemsg(f"mydep: {mydep}\n", 1)
         mydep = dep_expand(mydep, mydb=self.dbapi, settings=self.settings)
-        writemsg("mydep: %s\n" % mydep, 1)
+        writemsg(f"mydep: {mydep}\n", 1)
         mykey = dep_getkey(mydep)
-        writemsg("mykey: %s\n" % mykey, 1)
+        writemsg(f"mykey: {mykey}\n", 1)
         mymatch = best(match_from_list(mydep, self.dbapi.cp_list(mykey)))
-        writemsg("mymatch: %s\n" % mymatch, 1)
+        writemsg(f"mymatch: {mymatch}\n", 1)
         if mymatch is None:
             return ""
         return mymatch
@@ -2059,19 +2217,13 @@ class binarytree:
             elif binpkg_format == "xpak":
                 if self._multi_instance:
                     pf = catsplit(cpv)[1]
-                    filename = "{}-{}.xpak".format(
-                        os.path.join(self.pkgdir, cpv.cp, pf),
-                        "1",
-                    )
+                    filename = f"{os.path.join(self.pkgdir, cpv.cp, pf)}-1.xpak"
                 else:
                     filename = os.path.join(self.pkgdir, cpv + ".tbz2")
             elif binpkg_format == "gpkg":
                 if self._multi_instance:
                     pf = catsplit(cpv)[1]
-                    filename = "{}-{}.gpkg.tar".format(
-                        os.path.join(self.pkgdir, cpv.cp, pf),
-                        "1",
-                    )
+                    filename = f"{os.path.join(self.pkgdir, cpv.cp, pf)}-1.gpkg.tar"
                 else:
                     filename = os.path.join(self.pkgdir, cpv + ".gpkg.tar")
             else:
@@ -2118,17 +2270,6 @@ class binarytree:
             raise InvalidBinaryPackageFormat(binpkg_format)
 
     def _allocate_filename_multi(self, cpv, remote_binpkg_format=None):
-
-        # First, get the max build_id found when _populate was
-        # called.
-        max_build_id = self._max_build_id(cpv)
-
-        # A new package may have been added concurrently since the
-        # last _populate call, so use increment build_id until
-        # we locate an unused id.
-        pf = catsplit(cpv)[1]
-        build_id = max_build_id + 1
-
         if remote_binpkg_format is None:
             try:
                 binpkg_format = get_binpkg_format(cpv._metadata["PATH"])
@@ -2146,6 +2287,33 @@ class binarytree:
         else:
             raise InvalidBinaryPackageFormat(binpkg_format)
 
+        # If the preferred path is available then return
+        # that. This prevents unnecessary build_id incrementation
+        # triggered when the _max_build_id method counts remote
+        # build ids.
+        pf = catsplit(cpv)[1]
+        if getattr(cpv, "build_id", False):
+            preferred_path = f"{os.path.join(self.pkgdir, cpv.cp, pf)}-{cpv.build_id}.{binpkg_suffix}"
+            if not os.path.exists(preferred_path):
+                try:
+                    # Avoid races
+                    ensure_dirs(os.path.dirname(preferred_path))
+                    with open(preferred_path, "x") as f:
+                        pass
+                except FileExistsError:
+                    pass
+                else:
+                    return (preferred_path, cpv.build_id)
+
+        # First, get the max build_id found when _populate was
+        # called.
+        max_build_id = self._max_build_id(cpv)
+
+        # A new package may have been added concurrently since the
+        # last _populate call, so use increment build_id until
+        # we locate an unused id.
+        build_id = max_build_id + 1
+
         while True:
             filename = (
                 f"{os.path.join(self.pkgdir, cpv.cp, pf)}-{build_id}.{binpkg_suffix}"
@@ -2268,7 +2436,6 @@ class binarytree:
         return pkgindex
 
     def _get_digests(self, pkg):
-
         try:
             cpv = pkg.cpv
         except AttributeError:
