@@ -1,6 +1,7 @@
-# Copyright 2019-2021 Gentoo Authors
+# Copyright 2019-2024 Gentoo Authors
 # Distributed under the terms of the GNU General Public License v2
 
+import asyncio
 import functools
 import shutil
 import socket
@@ -10,8 +11,9 @@ import time
 
 import portage
 from portage.tests import TestCase
-from portage.util._eventloop.global_event_loop import global_event_loop
 from portage.util import socks5
+from portage.util.futures.executor.fork import ForkExecutor
+from portage.util._eventloop.global_event_loop import global_event_loop
 from portage.const import PORTAGE_BIN_PATH
 
 from http.server import BaseHTTPRequestHandler, HTTPServer
@@ -88,18 +90,20 @@ class AsyncHTTPServerTestCase(TestCase):
             if f is not None:
                 f.close()
 
-    def test_http_server(self):
+    async def _test_http_server(self):
+        asyncio.run(self._test_http_server())
+
+    async def _test_http_server(self):
         host = "127.0.0.1"
         content = b"Hello World!\n"
         path = "/index.html"
-        loop = global_event_loop()
+
+        loop = asyncio.get_running_loop()
         for i in range(2):
             with AsyncHTTPServer(host, {path: content}, loop) as server:
                 for j in range(2):
-                    result = loop.run_until_complete(
-                        loop.run_in_executor(
-                            None, self._fetch_directly, host, server.server_port, path
-                        )
+                    result = await loop.run_in_executor(
+                        None, self._fetch_directly, host, server.server_port, path
                     )
                     self.assertEqual(result, content)
 
@@ -177,38 +181,84 @@ class Socks5ServerTestCase(TestCase):
             return f.read()
 
     def test_socks5_proxy(self):
+        asyncio.run(self._test_socks5_proxy())
 
-        loop = global_event_loop()
+    async def _test_socks5_proxy(self):
+        loop = asyncio.get_running_loop()
 
         host = "127.0.0.1"
         content = b"Hello World!"
         path = "/index.html"
         proxy = None
         tempdir = tempfile.mkdtemp()
+        previous_exithandlers = portage.process._exithandlers
 
         try:
+            portage.process._exithandlers = []
             with AsyncHTTPServer(host, {path: content}, loop) as server:
-
                 settings = {
                     "PORTAGE_TMPDIR": tempdir,
                     "PORTAGE_BIN_PATH": PORTAGE_BIN_PATH,
                 }
 
                 proxy = socks5.get_socks5_proxy(settings)
-                loop.run_until_complete(socks5.proxy.ready())
+                await socks5.proxy.ready()
 
-                result = loop.run_until_complete(
-                    loop.run_in_executor(
-                        None,
-                        self._fetch_via_proxy,
-                        proxy,
-                        host,
-                        server.server_port,
-                        path,
-                    )
+                result = await loop.run_in_executor(
+                    None,
+                    self._fetch_via_proxy,
+                    proxy,
+                    host,
+                    server.server_port,
+                    path,
                 )
 
                 self.assertEqual(result, content)
         finally:
-            socks5.proxy.stop()
+            try:
+                # Also run_exitfuncs to test atexit hook cleanup.
+                await socks5.proxy.stop()
+                self.assertNotEqual(portage.process._exithandlers, [])
+                portage.process.run_exitfuncs()
+                self.assertEqual(portage.process._exithandlers, [])
+            finally:
+                portage.process._exithandlers = previous_exithandlers
+                shutil.rmtree(tempdir)
+
+
+class Socks5ServerLoopCloseTestCase(TestCase):
+    """
+    For bug 925240, test that the socks5 proxy is automatically
+    terminated when the main event loop is closed, using a subprocess
+    for isolation.
+    """
+
+    def testSocks5ServerLoopClose(self):
+        asyncio.run(self._testSocks5ServerLoopClose())
+
+    async def _testSocks5ServerLoopClose(self):
+        loop = asyncio.get_running_loop()
+        self.assertEqual(
+            await loop.run_in_executor(
+                ForkExecutor(loop=loop), self._testSocks5ServerLoopCloseSubprocess
+            ),
+            True,
+        )
+
+    @staticmethod
+    def _testSocks5ServerLoopCloseSubprocess():
+        loop = global_event_loop()
+        tempdir = tempfile.mkdtemp()
+        try:
+            settings = {
+                "PORTAGE_TMPDIR": tempdir,
+                "PORTAGE_BIN_PATH": PORTAGE_BIN_PATH,
+            }
+
+            socks5.get_socks5_proxy(settings)
+            loop.run_until_complete(socks5.proxy.ready())
+        finally:
+            loop.close()
             shutil.rmtree(tempdir)
+
+        return not socks5.proxy.is_running()
