@@ -1,13 +1,14 @@
-# Copyright 2012-2023 Gentoo Authors
+# Copyright 2012-2024 Gentoo Authors
 # Distributed under the terms of the GNU General Public License v2
 
 import fcntl
-import functools
 import multiprocessing
 import warnings
 import signal
 import sys
 
+from typing import Optional
+
 import portage
 from portage import os
 from portage.cache.mappings import slot_dict_class
@@ -23,17 +24,13 @@ class ForkProcess(SpawnProcess):
         "kwargs",
         "target",
         "_child_connection",
-        "_proc",
-        "_proc_join_task",
+        # Duplicate file descriptors for use by _send_fd_pipes background thread.
+        "_fd_pipes",
     )
 
     _file_names = ("connection", "slave_fd")
     _files_dict = slot_dict_class(_file_names, prefix="")
 
-    # Number of seconds between poll attempts for process exit status
-    # (after the sentinel has become ready).
-    _proc_join_interval = 0.1
-
     _HAVE_SEND_HANDLE = getattr(multiprocessing.reduction, "HAVE_SEND_HANDLE", False)
 
     def _start(self):
@@ -58,9 +55,14 @@ class ForkProcess(SpawnProcess):
                 duplex=self._HAVE_SEND_HANDLE
             )
 
-        retval = self._spawn(self.args, fd_pipes=self.fd_pipes)
+        # Handle fd_pipes in _main instead, since file descriptors are
+        # not inherited with the multiprocessing "spawn" start method.
+        # Pass fd_pipes=None to spawn here so that it doesn't leave
+        # a closed stdin duplicate in fd_pipes (that would trigger
+        # "Bad file descriptor" error if we tried to send it via
+        # send_handle).
+        self._proc = self._spawn(self.args, fd_pipes=None)
 
-        self.pid = retval[0]
         self._registered = True
 
         if self._child_connection is None:
@@ -73,13 +75,49 @@ class ForkProcess(SpawnProcess):
                 self.fd_pipes.setdefault(0, portage._get_stdin().fileno())
                 self.fd_pipes.setdefault(1, sys.__stdout__.fileno())
                 self.fd_pipes.setdefault(2, sys.__stderr__.fileno())
-                stdout_fd = os.dup(self.fd_pipes[1])
+                if self.create_pipe is not False:
+                    stdout_fd = os.dup(self.fd_pipes[1])
 
             if self._HAVE_SEND_HANDLE:
-                master_fd, slave_fd = self._pipe(self.fd_pipes)
-                self.fd_pipes[1] = slave_fd
-                self.fd_pipes[2] = slave_fd
+                if self.create_pipe is not False:
+                    master_fd, slave_fd = self._pipe(self.fd_pipes)
+                    self.fd_pipes[1] = slave_fd
+                    self.fd_pipes[2] = slave_fd
+                else:
+                    if self.logfile:
+                        raise NotImplementedError(
+                            "logfile conflicts with create_pipe=False"
+                        )
+                    # When called via process.spawn, SpawnProcess
+                    # will have created a pipe earlier, so it would be
+                    # redundant to do it here (it could also trigger spawn
+                    # recursion via set_term_size as in bug 923750). Use
+                    # /dev/null for master_fd, triggering early return
+                    # of _main, followed by _async_waitpid.
+                    # TODO: Optimize away the need for master_fd here.
+                    master_fd = os.open(os.devnull, os.O_RDONLY)
+                    slave_fd = None
+
                 self._files = self._files_dict(connection=connection, slave_fd=slave_fd)
+
+                # Create duplicate file descriptors in self._fd_pipes
+                # so that the caller is free to manage the lifecycle
+                # of the original fd_pipes.
+                self._fd_pipes = {}
+                fd_map = {}
+                for dest, src in list(self.fd_pipes.items()):
+                    if src not in fd_map:
+                        src_new = fd_map[src] = os.dup(src)
+                        old_fdflags = fcntl.fcntl(src, fcntl.F_GETFD)
+                        fcntl.fcntl(src_new, fcntl.F_SETFD, old_fdflags)
+                        os.set_inheritable(
+                            src_new, not bool(old_fdflags & fcntl.FD_CLOEXEC)
+                        )
+                    self._fd_pipes[dest] = fd_map[src]
+
+                asyncio.ensure_future(
+                    self._proc.wait(), self.scheduler
+                ).add_done_callback(self._close_fd_pipes)
             else:
                 master_fd = connection
 
@@ -87,6 +125,19 @@ class ForkProcess(SpawnProcess):
                 master_fd, log_file_path=self.logfile, stdout_fd=stdout_fd
             )
 
+    def _close_fd_pipes(self, future):
+        """
+        Cleanup self._fd_pipes if needed, since _send_fd_pipes could
+        have been cancelled.
+        """
+        # future.result() raises asyncio.CancelledError if
+        # future.cancelled(), but that should not happen.
+        future.result()
+        if self._fd_pipes is not None:
+            for fd in set(self._fd_pipes.values()):
+                os.close(fd)
+            self._fd_pipes = None
+
     @property
     def _fd_pipes_send_handle(self):
         """Returns True if we have a connection to implement fd_pipes via send_handle."""
@@ -101,16 +152,30 @@ class ForkProcess(SpawnProcess):
         Communicate with _bootstrap to send fd_pipes via send_handle.
         This performs blocking IO, intended for invocation via run_in_executor.
         """
-        fd_list = list(set(self.fd_pipes.values()))
-        self._files.connection.send(
-            (self.fd_pipes, fd_list),
-        )
-        for fd in fd_list:
-            multiprocessing.reduction.send_handle(
-                self._files.connection,
-                fd,
-                self.pid,
+        fd_list = list(set(self._fd_pipes.values()))
+        try:
+            self._files.connection.send(
+                (self._fd_pipes, fd_list),
             )
+            for fd in fd_list:
+                multiprocessing.reduction.send_handle(
+                    self._files.connection,
+                    fd,
+                    self.pid,
+                )
+        except BrokenPipeError as e:
+            # This case is triggered by testAsynchronousLockWaitCancel
+            # when the test case terminates the child process while
+            # this thread is still sending the fd_pipes (bug 923852).
+            # Even if the child terminated abnormally, then there is
+            # no harm in suppressing the exception here, since the
+            # child error should have gone to stderr.
+            raise asyncio.CancelledError from e
+
+        # self._fd_pipes contains duplicates that must be closed.
+        for fd in fd_list:
+            os.close(fd)
+        self._fd_pipes = None
 
     async def _main(self, build_logger, pipe_logger, loop=None):
         try:
@@ -128,12 +193,15 @@ class ForkProcess(SpawnProcess):
                     self._files.connection.close()
                     del self._files.connection
                 if hasattr(self._files, "slave_fd"):
-                    os.close(self._files.slave_fd)
+                    if self._files.slave_fd is not None:
+                        os.close(self._files.slave_fd)
                     del self._files.slave_fd
 
         await super()._main(build_logger, pipe_logger, loop=loop)
 
-    def _spawn(self, args, fd_pipes=None, **kwargs):
+    def _spawn(
+        self, args: list[str], fd_pipes: Optional[dict[int, int]] = None, **kwargs
+    ) -> portage.process.MultiprocessingProcess:
         """
         Override SpawnProcess._spawn to fork a subprocess that calls
         self._run(). This uses multiprocessing.Process in order to leverage
@@ -171,11 +239,7 @@ class ForkProcess(SpawnProcess):
                 )
                 fd_pipes[0] = stdin_dup
 
-            if self._fd_pipes_send_handle:
-                # Handle fd_pipes in _main instead.
-                fd_pipes = None
-
-            self._proc = multiprocessing.Process(
+            proc = multiprocessing.Process(
                 target=self._bootstrap,
                 args=(
                     self._child_connection,
@@ -186,19 +250,12 @@ class ForkProcess(SpawnProcess):
                     kwargs,
                 ),
             )
-            self._proc.start()
+            proc.start()
         finally:
             if stdin_dup is not None:
                 os.close(stdin_dup)
 
-        self._proc_join_task = asyncio.ensure_future(
-            self._proc_join(self._proc, loop=self.scheduler), loop=self.scheduler
-        )
-        self._proc_join_task.add_done_callback(
-            functools.partial(self._proc_join_done, self._proc)
-        )
-
-        return [self._proc.pid]
+        return portage.process.MultiprocessingProcess(proc)
 
     def _cancel(self):
         if self._proc is None:
@@ -206,64 +263,10 @@ class ForkProcess(SpawnProcess):
         else:
             self._proc.terminate()
 
-    def _async_wait(self):
-        if self._proc_join_task is None:
-            super()._async_wait()
-
-    def _async_waitpid(self):
-        if self._proc_join_task is None:
-            super()._async_waitpid()
-
-    async def _proc_join(self, proc, loop=None):
-        sentinel_reader = self.scheduler.create_future()
-        self.scheduler.add_reader(
-            proc.sentinel,
-            lambda: sentinel_reader.done() or sentinel_reader.set_result(None),
-        )
-        try:
-            await sentinel_reader
-        finally:
-            # If multiprocessing.Process supports the close method, then
-            # access to proc.sentinel will raise ValueError if the
-            # sentinel has been closed. In this case it's not safe to call
-            # remove_reader, since the file descriptor may have been closed
-            # and then reallocated to a concurrent coroutine. When the
-            # close method is not supported, proc.sentinel remains open
-            # until proc's finalizer is called.
-            try:
-                self.scheduler.remove_reader(proc.sentinel)
-            except ValueError:
-                pass
-
-        # Now that proc.sentinel is ready, poll until process exit
-        # status has become available.
-        while True:
-            proc.join(0)
-            if proc.exitcode is not None:
-                break
-            await asyncio.sleep(self._proc_join_interval, loop=loop)
-
-    def _proc_join_done(self, proc, future):
-        future.cancelled() or future.result()
-        self._was_cancelled()
-        if self.returncode is None:
-            self.returncode = proc.exitcode
-
-        self._proc = None
-        if hasattr(proc, "close"):
-            proc.close()
-        self._proc_join_task = None
-        self._async_wait()
-
     def _unregister(self):
         super()._unregister()
         if self._proc is not None:
-            if self._proc.is_alive():
-                self._proc.terminate()
-            self._proc = None
-        if self._proc_join_task is not None:
-            self._proc_join_task.cancel()
-            self._proc_join_task = None
+            self._proc.terminate()
 
     @staticmethod
     def _bootstrap(child_connection, have_send_handle, fd_pipes, target, args, kwargs):
